# -*- encoding: utf-8 -*-

__author__ = "Open Observatory of Network Interference"
__version__ = "1.6.0"
# This is the version number of resources to be downloaded
# when a release is made it should be aligned to __version__
__resources_version__ = "1.6.0"

__all__ = [
    'common',
    'deckgen',
    'nettest',
    'report',
    'resources',
    'templates',
    'utils'
]
