"""
This modules contains functionality that is shared amongst ooni-probe and
ooni-backend. If the code in here starts growing too much I think it would
make sense to either:

 * Make the code in here into it's own package that is imported by
 ooni-probe and ooni-backend.

 * Merge ooniprobe with oonibackend.
"""
